import copy
import numpy as numpy
from server.lib import hdf
from server.lib import convert as conv
from server.lib import votable as vot


class TreeManager(object):
    """
    Class returning a data structure conforming to the fancytre library
    https://wwwendt.de/tech/fancytree/demo/index.html#

    """
    def __init__(self, hdffile):
        self._tree = [{'title': '', 'key': 1, 'folder': True, 'children': []}]
        self._cache = {}  # cache for already explored paths
        self._nodecount = 1
        self._pointers = hdffile.index
        self._metadata = hdffile.metadata
        self._separator = hdffile.separator
        self._fillTree()
        
    @property
    def tree(self):
        return self._tree[0]
        
    def _incremNodeCount(self):
        self._nodecount = self._nodecount + 1

    def isCached(self, path):
        """
          Returns True is a path is in the cache
          @param path   path to test
        """
        if path in self._cache:
            return True
        return False

    def _getTooltip(self, pointed):
        """
          Returns the description of a pointer object to be used as a tooltip
          @param pointed a string used as a key to search into pointers map
          @param pointers a map of apointer objects
        """
        try:
            return self._pointers[pointed].description.replace("<![CDATA[", "").replace("]]>", "")
        except KeyError:
            return ""
        except AttributeError:
            return ""

    def _getDataType(self, branch):
        if "/Local quantities" in branch:
            return "local"
        return "integrated"

    def _recursTree(self, values, tree, path):
        """
          recurse on values until the leaf and append the lead to the list

        """
        local_list = copy.deepcopy(values)
        # search part of path already created in tree
        for value in local_list:
            for element in tree:
                if element['title'] == value:
                    local_list.pop(0)
                    if len(local_list) > 0:
                        self._recursTree(local_list, element["children"],
                                         path)
                    return True

        # current element still contains element, is a folder and has children
        if len(local_list) > 1:
            self._incremNodeCount()
            children = []
            tree.append({"title": local_list[0],
                         "key": self._nodecount,
                         "folder": True,
                         "children": children})
            local_list.pop(0)
            self._recursTree(local_list, children, path)
        # element is a leaf
        else:
            self._incremNodeCount()
            # quantity names are unique in data, used as key for tree leaves
            tree.append({"title": local_list[0],
                         "key": local_list[0],
                         "type": self._getDataType(path),
                         "tooltip": self._getTooltip(local_list[0])})
            self._cache[path] = tree

    def _addValue(self, line_utf):
        """
          add content of values list into path
          each element of the list will be a new node in the tree if
          it does not exist already
          the content of path variable is used as an index in the cache
          to directly get the list of leaves at this position

          @param values list describing path in the tree
          @param path string containing path in the tree without the final leaf
        """
        
        path = line_utf[0] + "/" + line_utf[1]
        line = (path).replace("/", self._separator)\
                                  .split(self._separator)
        line.append(line_utf[4])

        # if a path is already cached, append the node to it
        if path in self._cache:
            self._incremNodeCount()
            self._cache[path].append({"title": line[len(line) - 1],
                                     "unit": line_utf[6],
                                     "key": line[len(line) - 1],
                                     "type": self._getDataType(path)})
        else:
            local_list = copy.deepcopy(line)
            for value in local_list:
                for element in self._tree:
                    if element['title'] == value:
                        local_list.pop(0)
                        self._recursTree(local_list,
                                         element["children"],
                                         path)
                        return True

            self._incremNodeCount()
            self._tree.append({"title": local_list[0],
                              "key": self._nodecount,
                              "unit": line_utf[6],
                              "folder": True,
                              "children": []})

    def _fillTree(self):
        for line in self._metadata:
            line_utf = []
            # cast everything to string to avoid problems in gui
            # when comparing used input and index content
            # performance pust be checked on big files
            for value in line:
                line_utf.append(value)

            self._addValue(line_utf)


class Writer(object):
    """
      write extracted data in a file
    """
    def __init__(self, filename, header=None, separator=',', isdouble=False):
        # output file name
        self.filename = filename
        # a list of pointers
        self.pointers = []
        # a list of columns
        self.columns = []
        self.headerColumns = []
        # separator between values
        self.separator = separator
        # precision is .6 if False, .15 if True
        self.isDouble = isdouble
        self.header = header

    def addColumn(self, pointer, column):
        """
            add a column of data and its pointer
        """
        self.pointers.append(pointer)
        self.columns.append(column)

    def clearColumns(self):
        self.pointers = []
        self.columns = []

    def _getEmptyData(self, length):
        result = []
        for i in range(0, length):
            result.append('Not_In_HDF5')
        return result

    def _getArraySize(self):
        length = 0
        for i in range(0, len(self.columns)):
            if len(self.columns[i]) > 0:
                length = len(self.columns[i])
                break
        return length

    def writeText(self, replace_ws=False):
        """
            write all data
        """
        try:
            with open(self.filename, 'w') as f:
                precision = "float"
                self._columnsAreWritable()
                data = []
                widths = []
                length = self._getArraySize()
                
                for entry in self.header.values():
                    f.write("# %s : %s \n" % (entry['label'], entry['value']))
                for i in range(len(self.columns)):
                    if len(self.columns[i]) == 0:
                        self.columns[i] = self._getEmptyData(length)
                
                for i in range(0, length):
                    values = [column[i] for column in self.columns]
                    towrite = []
                    for value in values:
                        if not isinstance(value, str):
                            value = conv.getValue(value, precision)
                        widths.append(len(value))
                        towrite.append(value.strip())
                    data.append(self.separator.join(towrite))
                if length > 1:
                    # common case, several lines, several columns
                    line1 = []
                    line2 = []
                    line3 = []
                    for idx, pointer in enumerate(self.pointers):
                        if replace_ws:
                            line1.append('[' + str(idx + 1) + ']' +
                                         pointer.name.replace(" ", replace_ws))
                            line2.append('[' + str(idx + 1) + ']' +
                                         pointer.unit.replace(' ', replace_ws))
                        else:
                            line1.append('[' + str(idx + 1) + ']' + pointer.name)
                            line2.append('[' + str(idx + 1) + ']' + pointer.unit)

                        line3.append(('[' + str(idx + 1) + ']').rjust(widths[idx]))
                        header = "#" + self.separator.join(line1) + "\n#" + \
                                 self.separator.join(line2) + "\n#" + \
                                 self.separator.join(line3)[1:] + "\n"

                    f.write(header)
                    f.write("\n".join(data))
                else:
                    # file with only one line and several columns
                    # are plotted as one column with several lines
                    f.write("#%s%s       %s    %s\n" % ("index".rjust(6),
                                                        "value".rjust(35),
                                                        "unit".rjust(12),
                                                        "quantity"))
                    for idx, value in enumerate(data[0].split(self.separator)):
                        f.write("%s%s   #   %s     %s\n" % (str(idx).rjust(7),
                                                            str(value).rjust(35),
                                                            self.pointers[idx].unit.rjust(11),
                                                            self.pointers[idx].name))
        except Exception as e:
            raise e

    def writeVoTable(self):
        try:
            self._columnsAreWritable()
            table = vot.Votable()
            datatype = 'float'
            builder = vot.InfoBuilder()
            for key, value in self.header.items():
                table.addInfo(builder.withName(value['label'])
                                     .withValue(value['value'])
                                     .getInfo())
                builder.reset()
            builder = vot.FieldBuilder()
            for i in range(0, len(self.columns)):
                field = self.pointers[i]
                link = vot.LinkBuilder()\
                          .withContentRole('type')\
                          .withHref(field.skos)\
                          .getLink()
                table.addField(builder.withName(field.name)
                                      .withUnit(field.unit)
                                      .withUcd(field.ucd)
                                      .withUtype(field.utype)
                                      .withDatatype(datatype)
                                      .withLink(link)
                                      .getField())
                table.addColumn(self.columns[i])
                builder.reset()
            table.toFile(self.filename)

        except Exception as e:
            raise e

    def _columnsAreWritable(self):
        """
            check data validity
        """
        if len(self.columns) != len(self.pointers):
            raise Exception("columns and pointers do not match")

        if len(self.columns) > 0:
            size = self._getArraySize()
            for i in range(1, len(self.columns)):
                if len(self.columns[i]) > 0 and len(self.columns[i]) != size:
                    raise Exception('all the columns do not have the same size')
            return True
        raise Exception("no data")


class Exporter(object):

    @staticmethod
    def _updateHeader(hdf5, header):
        try:
            # _c_ files contains the code version
            header["code version"] = {"value": hdf5.getCode(),
                                      "label": "Produced by"}
        except KeyError as e:
            # _s_ files does not contain the information
            header["code version"] = {  "value": "unknown", 
                                        "label": "Produced by"}
            
            

    @staticmethod
    def exportAsText(hdf5, header,
                     columns, outputfile,
                     separator="   ", isdouble=False):
        """
            export selected data into text file
        """
        factory = hdf.PointerFactory()
        w = Writer(outputfile, header, separator=separator, isdouble=isdouble)
        # add code version to header
        Exporter._updateHeader(hdf5, header)
        for i in range(0, len(columns)):
            try:
                column = columns[i]                
                w.addColumn(hdf5.index[column],
                            hdf5.extractQuantity(hdf5.index[column]))
            except Exception as e:
                print(e)
                pointer = factory.getErrorPointer(column)
                w.addColumn(pointer, [])
            try:
                w.writeText()
            except Exception as e:
                print(e)
                raise e

    @staticmethod
    def exportAsVotable(hdf5, header, columns, outputfile, isdouble=False):
        """
            export selected data into text file
        """
        w = Writer(outputfile, header, isdouble=isdouble)
        # add code version to header
        Exporter._updateHeader(hdf5, header)
        rejected = []
        for i in range(0, len(columns)):
            try:
                column = columns[i]
                w.addColumn(hdf5.index[column],
                            hdf5.extractQuantity(hdf5.index[column]))
            except Exception as e:
                rejected.append(column)

        if len(rejected) > 0:
            Exporter.exportErrors(outputfile, rejected)

        try:
            w.writeVoTable()
        except Exception as e:
            raise e

    @staticmethod
    def exportErrors(outputfile, errors):
        parts = outputfile.split('.')
        filepath = parts[0] + '_errors.' + parts[1]
        with open(filepath, 'w') as f:
            f.write("unknown columns : \n")
            for error in errors:
                f.write(error + '\n')
