import os
import os.path
import sys
import json
import uuid
import time
from datetime import datetime
import traceback
import codecs

from server.lib import hdf
from server.lib import hdfutil
from server.lib import database as jobdb
from server.lib import config as conf
from server.lib import logger as logger
from idatplot import plot as pdrplot


def export(job_id, job_config, db_int,
           export_function, header, output_filename):
    is_completed = True
    # returns a (head, tail) tuple
    path = os.path.split(output_filename)
    if path[0] == '':
        output_path = conf.OUTPUT_DIRECTORY
    else:
        output_path = os.path.join(conf.OUTPUT_DIRECTORY, path[0])
        try:
            os.mkdir(output_path)
        except IOError:
            # directory already exists
            pass
    try:
        hdffile = hdf.PdrHDF(os.path.join(conf.DATA_DIRECTORY,
                             job_config['hdffile']))
        db_int.change_job_status(job_id, "EXECUTING")
        export_function(hdffile, header,
                        job_config['quantities'],
                        os.path.join(output_path,
                                     path[1]))
    except Exception as e:
        traceback.print_exc()
        is_completed = False
        db_int.set_error_status(job_id, str(e))

    if is_completed is True:
        db_int.set_completed_status(job_id, output_filename)

def append_extension(path, extension):
    if extension not in path :
        return path + extension
    else :
        return path


def execute_extract_job(db_int, job):
    job_config = json.loads(job['parameters'])
    if job_config["output_path"] is not None:
        output_filename = job_config["output_path"]
    else:
        output_filename = str(uuid.uuid4())

    header = {"hdffile":
              {"value": job_config['hdffile'],
               "label": "Model name"},
              "date": {"value": datetime.now().strftime('%d-%m-%Y'),
                       "label": "Extraction date"}
              }

    # if we are in web mode, hdffile is project/model and / is replaced
    # by a _ to build extract filename
    output_filename = output_filename + \
                      "_"+ \
                      job_config['hdffile'].replace(".hdf5", "").replace('/', '_') + \
                      "_" + \
                      job_config['datatype']
                      
    if job_config['dataformat'] == "text":
        output_filename = append_extension(output_filename, ".txt")
        f = hdfutil.Exporter.exportAsText
        export(job['id'], job_config, db_int,
               f, header, output_filename)
    elif job_config['dataformat'] == "votable":
        output_filename = append_extension(output_filename, ".xml")
        f = hdfutil.Exporter.exportAsVotable
        export(job['id'], job_config, db_int,
               f, header, output_filename)


def execute_plot_job(db_int, job): 
    """
        @param db_int database connection
        @param job job parameters
    """   
    try:        
        job_config = json.loads(job['parameters'])

        yaml_content = job_config['plotconfiguration']['yaml']
        output = job_config['output']
        yaml_file = os.path.join(conf.OUTPUT_DIRECTORY, output + ".yaml")
        yaml_file_nomodel = os.path.join(conf.OUTPUT_DIRECTORY, conf.YAML_TEMPLATE_NAME)
        hdf_file = os.path.join(conf.DATA_DIRECTORY, job_config['hdffile'])      

        yaml_str = codecs.decode(codecs.encode(yaml_content, 
                                                "utf-8", 
                                                'backslashreplace'),
                                 'unicode-escape')  

        # write yaml config file
        with open(yaml_file, 'w') as f:
            f.write(yaml_str)

        # write yaml config file without model name to use as templace
        with open(yaml_file_nomodel, 'w') as f:            
            f.write(yaml_str.replace(output, 'MODELNAME'))
            
        pdrplot.cmdline_client(yaml_file,
                               hdf_file,
                               conf.OUTPUT_DIRECTORY)
        db_int.set_completed_status(job['id'], output + '.png')
    except Exception as e:
        print(e)
        traceback.print_exc()
        db_int.set_error_status(job['id'], str(e))


def remove_file(path):
    try:
        os.remove(path)
    except IOError:
        pass


def delete_old_jobs(db_int, jobs, duration):
    for job in jobs:
        if time.time() - job['creation_time'] > duration:
            db_int.delete_job_by_id(job['id'])
            if job['result'] is not None:
                try:
                    uuid = job['result'].split('.')[0]
                    remove_file(os.path.join(conf.OUTPUT_DIRECTORY, job['result']))
                    remove_file(os.path.join(conf.OUTPUT_DIRECTORY, uuid + ".dat"))
                    remove_file(os.path.join(conf.OUTPUT_DIRECTORY, uuid + ".yaml"))
                    remove_file(os.path.join(conf.OUTPUT_DIRECTORY, uuid + ".txt"))
                except IOError as e:
                    traceback.print_exc()
                    logger.append_exception(e)
                    logger.append_info("File %s does not exist for job %s" %
                                       (job['result'], job['id']))


def clean_database(db_int):
    """
        Remove all content from database
    """
    db_int.purge_sessions()
    db_int.purge_joblists()
    db_int.purge_jobs()

def clean_directories():
    """
        Remove all content from data and output directories
    """
    for f in os.listdir(conf.OUTPUT_DIRECTORY):
        os.remove(os.path.join(conf.OUTPUT_DIRECTORY, f))

    if conf.SERVER_MODE == 'local':
        for f in os.listdir(conf.DATA_DIRECTORY):
            os.remove(os.path.join(conf.DATA_DIRECTORY, f))



def run():
    logger.append_info("Job processor service started")
    db_int = jobdb.JobsInterface()
    db_int.connect()
    clean_database(db_int)
    clean_directories()
    logger.append_info("Start listening loop")
    while True:
        time.sleep(conf.JOB_CHECK_INTERVAL)
        try:
            job_to_run = db_int.get_next_job()
            if job_to_run is not None:
                if job_to_run['type'] == 'EXTRACT':
                    execute_extract_job(db_int, job_to_run)
                elif job_to_run['type'] == 'PLOT':
                    execute_plot_job(db_int, job_to_run)
        except Exception as e:
            print(e)
            traceback.print_exc()
            db_int.disconnect()
            sys.exit()


if __name__ == "__main__":
    run()
